########################################################################################################################
# IMPORTS

import logging

from pydrive2.auth import GoogleAuth
from pydrive2.drive import GoogleDrive

########################################################################################################################
# CLASSES

logger = logging.getLogger(__name__)
logging.getLogger("googleapicliet.discovery_cache").setLevel(logging.ERROR)


class DriveInterface:
    def __init__(self, config):
        if "drive" in config:
            self.config = config["drive"]

            GoogleAuth.DEFAULT_SETTINGS[
                "client_config_file"
            ] = f'{self.config["config_path"]}/credentials.json'

            self.gauth = GoogleAuth(
                settings_file=f'{self.config["config_path"]}/settings.yaml'
            )
            self.gauth.LocalWebserverAuth()

            self.drive = GoogleDrive(self.gauth)

            self.team_id = self.config["team_id"]

        else:
            logger.warning("no drive section in config")

    def delete_old_files(self, filename, folder_id):
        for drive_file in self.drive.ListFile(
            {
                "q": f"'{folder_id}' in parents and trashed=false",
                "corpora": "teamDrive",
                "teamDriveId": self.team_id,
                "includeTeamDriveItems": True,
                "supportsTeamDrives": True,
            }
        ).GetList():
            if drive_file["title"] == filename:
                logger.info(f"deleting old {filename}...")
                drive_file.Delete(param={"supportsTeamDrives": True})

    def upload_file(self, local_filename, drive_filename, folder_id):
        self.delete_old_files(drive_filename, folder_id)

        f = self.drive.CreateFile(
            {
                "title": drive_filename,
                "parents": [
                    {
                        "kind": "drive#fileLink",
                        "teamDriveId": self.team_id,
                        "id": folder_id,
                    }
                ],
            }
        )
        f.SetContentFile(local_filename)

        logger.info(f"uploading {drive_filename} to folder: {folder_id}...")
        f.Upload(param={"supportsTeamDrives": True})

    def validate_file(self, filename, folder_id):
        for drive_file in self.drive.ListFile(
            {
                "q": f"'{folder_id}' in parents and trashed=false",
                "corpora": "teamDrive",
                "teamDriveId": self.team_id,
                "includeTeamDriveItems": True,
                "supportsTeamDrives": True,
            }
        ).GetList():
            if drive_file["title"] == filename:
                logger.info(f"{filename} correctly uploaded.")
                return

        raise FileNotFoundError(f"{filename} has not been correctly uploaded.")



########################################################################################################################
# IMPORTS

import logging
import time
import random

import requests
from stem import Signal
from stem.control import Controller

########################################################################################################################
# CLASSES

logger = logging.getLogger(__name__)
logging.getLogger("stem").setLevel(logging.WARNING)


class ProxyInterface:
    CHECK_IP_URL = "https://wtfismyip.com/json"

    def __init__(self, config):
        if "proxy" in config:
            self.config = config["proxy"]
            self.current_index = 0  # Initialize index for round robin
        else:
            logger.warning("no proxy section in config")

    @property
    def proxies(self):
        return self.get_proxies(use_tor="tor_password" in self.config)

    @staticmethod
    def get_proxy_url(host, port, user=None, password=None, use_socks=True):
        proxy_url = f"{host}:{port}"

        if user and password:
            proxy_url = f"{user}:{password}@{proxy_url}"

        proxy_url = f"socks5://{proxy_url}" if use_socks else f"http://{proxy_url}"
        return proxy_url

    def get_proxies(self, use_tor=False, randomize=False):
        if use_tor:
            proxy_url = self.get_proxy_url("127.0.0.1", 9050)
        else:
            current_host, current_port = self.get_random_host_port() if randomize else self.get_current_host_port()
            
            user = self.config.get("user")
            password = self.config.get("password")
            use_socks = self.config.get("socks", "false").lower() == "true"

            proxy_url = self.get_proxy_url(current_host, current_port, user, password, use_socks)

        return {
            "http": proxy_url,
            "https": proxy_url,
        }

    def get_current_host_port(self):
        hosts = self.config["hosts"].split(",") if isinstance(self.config["hosts"], str) else self.config["hosts"]
        host_port_pairs = [hp.split(":") for hp in hosts]
        current_host, current_port = host_port_pairs[self.current_index]
        self.current_index = (self.current_index + 1) % len(host_port_pairs)
        return current_host, current_port

    def get_random_host_port(self):
        hosts = self.config["hosts"].split(",") if isinstance(self.config["hosts"], str) else self.config["hosts"]
        host_port_pairs = [hp.split(":") for hp in hosts]
        self.current_index = random.randint(0, len(host_port_pairs) - 1)
        return host_port_pairs[self.current_index]

    def check_current_ip(self):
        try:
            return requests.get(self.CHECK_IP_URL, proxies=self.proxies).json()[
                "YourFuckingIPAddress"
            ]
        except Exception as ex:
            logger.error(ex)

    def renew_tor_ip(self):
        try:
            logger.info(f"renewing Tor ip: {self.check_current_ip()}...")
            with Controller.from_port(port=9051) as controller:
                controller.authenticate(password=self.config["tor_password"])
                controller.signal(Signal.NEWNYM)

            time.sleep(5)
            logger.info(f"new Tor IP: {self.check_current_ip()}")

        except Exception as ex:
            logger.error("unable to renew Tor ip")
            logger.error(ex)



########################################################################################################################
# IMPORTS

import io
import logging
import boto3

########################################################################################################################
# CLASSES

logger = logging.getLogger(__name__)


class AWSInterface:
    def __init__(self, config):
        self.profiles = []
        self.config = config

        for section in self.config.sections():
            if section.startswith("aws:"):
                profile_name = section.split(":", 1)[1]
                self.profiles.append(
                    {
                        "profile": profile_name,
                        "bucket": self.config[section]["bucket"],
                        "session": boto3.Session(profile_name=profile_name),
                    }
                )

        if not self.profiles:
            logger.warning("No AWS profiles found in config file")

        self.current_profile = self.profiles[0] if self.profiles else None
        self._update_resources()

    def _update_resources(self):
        if self.current_profile:
            self.s3 = self.current_profile["session"].resource("s3")
            self.s3_client = self.s3.meta.client
            self.bucket = self.current_profile["bucket"]

    def switch_profile(self, profile_name):
        for profile in self.profiles:
            if profile["profile"] == profile_name:
                self.current_profile = profile
                self._update_resources()
                return
        logger.warning(f"Profile {profile_name} not found")

    def get_file(self, s3_path):
        try:
            return self.s3.Object(self.bucket, s3_path).get()
        except self.s3_client.exceptions.NoSuchKey:
            logger.info(f"{s3_path} does not exist")

    def read_file_as_bytes(self, s3_path):
        return io.BytesIO(self.get_file(s3_path)["Body"].read())

    def upload_file(self, local_path, s3_path):
        self.s3.Bucket(self.bucket).upload_file(local_path, s3_path)



########################################################################################################################
# IMPORTS

import logging
from urllib.parse import quote_plus

from sqlalchemy import DDL, create_engine, text
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker

########################################################################################################################
# CLASSES

logger = logging.getLogger(__name__)


class MockContext:
    def __init__(self, column):
        self.current_parameters = {}
        self.current_column = column
        self.connection = None


class AlchemyInterface:
    def __init__(self, config):
        if "db" in config:
            self.config = config["db"]

            self.engine = create_engine(self.get_conn_str())
            self.session = sessionmaker(bind=self.engine)()
            self.cursor = self.session.connection().connection.cursor()

        else:
            logger.warning("no db section in config")

    def get_conn_str(self):
        return (
            f"{self.config['engine']}://"
            f"{self.config['user']}:{quote_plus(self.config['password'])}"
            f"@{self.config['host']}:{self.config['port']}"
            f"/{self.config['database']}"
        )

    @staticmethod
    def get_schema_from_table(table):
        schema = "public"

        if isinstance(table.__table_args__, tuple):
            for table_arg in table.__table_args__:
                if isinstance(table_arg, dict) and "schema" in table_arg:
                    schema = table_arg["schema"]

        elif isinstance(table.__table_args__, dict) and "schema" in table.__table_args__:
            schema = table.__table_args__["schema"]

        if schema == "public":
            logger.warning(f"no database schema provided, switching to {schema}...")

        return schema

    def create_tables(self, tables):
        for table in tables:
            schema = self.get_schema_from_table(table)

            with self.engine.connect() as conn:
                conn.execute(DDL(f"CREATE SCHEMA IF NOT EXISTS {schema}"))
                conn.commit()

                if hasattr(table, "is_view") and table.is_view:
                    if not conn.dialect.has_table(conn, table.__tablename__, schema=schema):
                        logger.info(f"creating view {table.__tablename__}...")
                        table.create_view(conn)
                        conn.commit()
                    else:
                        logger.info(f"view {table.__tablename__} already exists")
                else:
                    if not conn.dialect.has_table(conn, table.__tablename__, schema=schema):
                        logger.info(f"creating table {table.__tablename__}...")
                        table.__table__.create(conn)
                        conn.commit()
                    else:
                        logger.info(f"table {table.__tablename__} already exists")

    def drop_tables(self, tables):
        for table in tables:
            schema = self.get_schema_from_table(table)

            with self.engine.connect() as conn:
                if hasattr(table, "is_view") and table.is_view:
                    if conn.dialect.has_table(conn, table.__tablename__, schema=schema):
                        logger.info(f"dropping view {table.__tablename__}...")
                        conn.execute(DDL(f"DROP VIEW {schema}.{table.__tablename__} CASCADE"))
                        conn.commit()
                else:
                    if conn.dialect.has_table(conn, table.__tablename__, schema=schema):
                        logger.info(f"dropping table {table.__tablename__}...")
                        conn.execute(DDL(f"DROP TABLE {schema}.{table.__tablename__} CASCADE"))
                        conn.commit()

    def reset_db(self, tables, drop):
        if drop:
            self.drop_tables(tables)

        self.create_tables(tables)

    def insert_alchemy_obj(self, alchemy_obj, silent=False):
        try:
            if not silent:
                logger.info(f"adding {alchemy_obj}...")

            self.session.add(alchemy_obj)
            self.session.commit()

        except IntegrityError:
            if not silent:
                logger.info(f"{alchemy_obj} already in db")

            self.session.rollback()

    def upsert_alchemy_obj(self, alchemy_obj, index_elements, silent=False):
        if not silent:
            logger.info(f"upserting {alchemy_obj}")

        primary_keys = list(col.name for col in alchemy_obj.__table__.primary_key.columns.values())
        obj_dict = {
            col.name: val
            for col in alchemy_obj.__table__.columns
            if col.name not in primary_keys and (val := getattr(alchemy_obj, col.name)) is not None
        }

        statement = (
            insert(alchemy_obj.__table__)
            .values(obj_dict)
            .on_conflict_do_update(index_elements=index_elements, set_=obj_dict)
        )

        try:
            self.session.execute(statement)
            self.session.commit()
        except IntegrityError:
            if not silent:
                logger.info(f"could not upsert {alchemy_obj}")

            self.session.rollback()

    def reset_column(self, query_results, column_name):
        if not query_results:
            logger.warning("No objects to reset column for.")
            return

        first_obj = query_results[0]
        model_class = first_obj.__class__
        table = model_class.__table__

        if column_name not in table.columns:
            logger.warning(f"Column {column_name} does not exist in table {table.name}.")
            return

        column = table.columns[column_name]

        if column.server_default is not None:
            query_results.update({column_name: text("DEFAULT")}, synchronize_session=False)
        elif column.default is not None:
            default_value = column.default.arg
            if callable(default_value):
                default_value = default_value(MockContext(column))
            query_results.update({column_name: default_value}, synchronize_session=False)
        else:
            raise ValueError(f"Column '{column_name}' doesn't have a default value defined.")

        self.session.commit()



########################################################################################################################
# IMPORTS

import logging

import requests

from ..params.nominatim import POSTCODES

########################################################################################################################
# CLASSES

logger = logging.getLogger(__name__)


class GeoNames:
    def __init__(self, endpoint):
        self.endpoint = endpoint

    @staticmethod
    def validate_postcode(postcode):
        if isinstance(postcode, int):
            postcode = str(postcode)

        if postcode and len(postcode) == 5 and postcode[:2] in POSTCODES:
            return postcode

        if postcode and len(postcode) == 4:
            postcode = f"0{postcode}"
            if postcode[:2] in POSTCODES:
                return postcode

    @staticmethod
    def get_province_from_postcode(postcode):
        if postcode:
            return POSTCODES[postcode[:2]]

    def reverse(self, lat, lon):
        return requests.get(f"{self.endpoint}/reverse?lat={lat}&lon={lon}").json()


class Nominatim:
    def __init__(self, nominatim_endpoint, geonames_endpoint):
        self.endpoint = nominatim_endpoint
        self.geonames = GeoNames(geonames_endpoint)

    @staticmethod
    def get_attribute(raw_json, keys):
        for key in keys:
            if key in raw_json:
                return raw_json[key]

    def geocode(self, address):
        return requests.get(f"{self.endpoint}/search?q={address}&format=json").json()

    def geocode_parsed(self, address):
        results = self.geocode(address)

        if results:
            return self.reverse_parsed(results[0]["lat"], results[0]["lon"])

    def reverse(self, lat, lon):
        return requests.get(f"{self.endpoint}/reverse?lat={lat}&lon={lon}&format=json").json()

    def reverse_parsed(self, lat, lon):
        raw_json = self.reverse(lat, lon).get("address", {})
        geoname = self.geonames.reverse(lat, lon)

        postcode = self.geonames.validate_postcode(
            str(geoname.get("postal_code", ""))
        ) or self.geonames.validate_postcode(str(raw_json.get("postcode")))

        city = self.get_attribute(raw_json, ["city", "town", "village"]) or geoname.get("place_name")

        district, quarter = self.get_district_quarter(raw_json)
        return {
            "country": raw_json.get("country"),
            "country_code": (raw_json.get("country_code") or geoname.get("country_code") or "").lower(),
            "state": raw_json.get("state") or geoname.get("community"),
            "province": self.geonames.get_province_from_postcode(postcode),
            "city": city,
            "postcode": postcode,
            "district": district,
            "quarter": quarter,
            "street": raw_json.get("road"),
            "number": raw_json.get("house_number"),
        }

    def get_district_quarter(self, raw_json):
        district = self.get_attribute(raw_json, ["city_district", "suburb", "borough"])
        quarter = self.get_attribute(raw_json, ["quarter", "neighbourhood"])

        if not district and quarter:
            district = quarter
            quarter = None

        return district, quarter


class NominatimInterface(Nominatim):
    def __init__(self, config):
        if "osm" in config:
            self.config = config["osm"]

            self.nominatim_endpoint = self.config["nominatim_endpoint"]
            self.geonames_endpoint = self.config["geonames_endpoint"]

            super().__init__(self.nominatim_endpoint, self.geonames_endpoint)
        else:
            logger.warning("no osm section in config")



########################################################################################################################
# IMPORTS

import json

import typer
from typing_extensions import Annotated

########################################################################################################################
# TYPES

class Dict(dict):
    def __init__(self, value: str):
        super().__init__(json.loads(value))

    def __repr__(self):
        return f"Dict({super().__repr__()})"


def parse_json_dict(value: str) -> Dict:
    try:
        return Dict(value)
    except json.JSONDecodeError as err:
        raise ValueError(f"Invalid JSON string: {value}") from err


DictArg = Annotated[Dict, typer.Argument(parser=parse_json_dict)]
DictOpt = Annotated[Dict, typer.Option(parser=parse_json_dict)]